{
Client Access Objects 1.5

caoQuery component

(c) 1998, 1999 Steven P. Goldsmith
All Rights Reserved

The caoQuery component implements a TDataSet descendant for accessing data on
the AS/400.

The PChar record used by TDataSet is formatted as follows:

------------------------------------------------------------
| Record Data | Null Indicators | Bookmark | Bookmark Flag |
------------------------------------------------------------

Features:

o  Smart record caching improves speed and reduces network traffic.

o  All AS/400 field types supported as of V4R3.

o  All data conversion is done inside the component for greatest speed and
   flexability.  This also eliminates the need to call the Client Access APIs.

Known limitations:

o  Currently read only access.  I've added the SetFieldData method that converts
   Delphi types to AS/400.  InternalPost needs to be implemented.

o  Date, Time and Timestamp are handled as strings since this is the way Client
   Access hands them back from GetResultData.

o  The following methods of TDataset were not used because it was hard to get a
   reliable record count from SQLCA value sqlerrd3.

  o  function GetRecordCount: Integer; override;
  o  function GetRecNo: Integer; override;
  o  procedure SetRecNo(Value: Integer); override;

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 1, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

sgoldsmith@home.com
http://members.home.net/sgoldsmith/
}

unit caoQuery;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Db, caoDatabaseSQL,

  // Import type library "Client Access Object Library" before installing.

  Cwbautsv_TLB;

type

  // caoQuery exception object

  EcaoQuery = class(Exception);

  // Controls what options are available on query open

  TcaoOpenOptions = (
    caodbRead,
    caodbWrite,
    caodbUpdate,
    caodbDelete,
    caodbOpenAll);

  // AS/400 specific field definition

  TcaoFieldDef = class(TCollectionItem)
  public
    ColName: String;
    ColType: SmallInt;
    ColLength: Integer;
    ColPerc: SmallInt;
    ColScale: SmallInt;
    ColCCSID: SmallInt;
    ColOffset: Integer;
  end;

  // Bookmark record

  PcaoRecInfo = ^TcaoRecInfo;
  TcaoRecInfo = packed record
    Bookmark: Integer;
    BookmarkFlag: TBookmarkFlag;
  end;

  TcaoQuery = class(TDataSet)
  private
    { Private declarations }
    FCursorOpen: Boolean;          // Is the cusror open?
    FRecBufSize: Integer;          // Size of record buffer
    FRecInfoOfs: Integer;          // Start of bookmark record
    FRecNullInfoOfs: Integer;      // Start of null indicators
    FCacheCount: Integer;          // caoDatabaseSQL.defaultBlockCount
    FCurRec: Integer;              // Current physical record number
    FCurCacheRec: Integer;         // Current physical record number in cache
    FFirstCacheRec: Integer;       // First physical record number in cache
    FLastCacheRec: Integer;        // Last physical record number in cache
    FRecCache: array of String;    // Dynamic record cache buffer
    FFieldDefs400: TCollection;    // AS/400 specific field definitions
    FstatementName: WideString;    // SQL statement name
    FstatementText: TStrings;      // SQL statement text
    FcursorName: WideString;       // Name of cursor
    FopenOptions: TcaoOpenOptions; // Cursor open mode
    FcaoDatabaseSQL: TcaoDatabaseSQL;
    function VarByteToStr(V: OleVariant): String;
    function VarSmallIntToStr(V: OleVariant): String;
  protected
    { Protected declarations }
    procedure SetLines(Value: TStrings); virtual;
    procedure Notification(AComponent: TComponent;
      Operation: TOperation); override;
    function GetFieldDefs400: Integer; virtual;
    function CacheResultData: Integer; virtual;
    procedure LastRecord; virtual;
    function NextCacheRec: TGetResult; virtual;
    function PriorCacheRec: TGetResult; virtual;
    // Overriden abstract methods (required)
    function GetCanModify: Boolean; override;
    function AllocRecordBuffer: PChar; override;
    procedure FreeRecordBuffer(var Buffer: PChar); override;
    procedure GetBookmarkData(Buffer: PChar; Data: Pointer); override;
    function GetBookmarkFlag(Buffer: PChar): TBookmarkFlag; override;
    function GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult; override;
    function GetRecordSize: Word; override;
    procedure InternalClose; override;
    procedure InternalFirst; override;
    procedure InternalGotoBookmark(Bookmark: Pointer); override;
    procedure InternalHandleException; override;
    procedure InternalInitFieldDefs; override;
    procedure InternalInitRecord(Buffer: PChar); override;
    procedure InternalLast; override;
    procedure InternalOpen; override;
    procedure InternalPost; override;
    procedure InternalSetToRecord(Buffer: PChar); override;
    function IsCursorOpen: Boolean; override;
    procedure SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag); override;
    procedure SetBookmarkData(Buffer: PChar; Data: Pointer); override;
    procedure SetFieldData(Field: TField; Buffer: Pointer); override;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function GetFieldData(Field: TField; Buffer: Pointer): Boolean; override;
    procedure ExecSQL; virtual;
  published
    { Published declarations }
    property Active;
    property BeforeOpen;
    property AfterOpen;
    property BeforeClose;
    property AfterClose;
    property BeforeInsert;
    property AfterInsert;
    property BeforeEdit;
    property AfterEdit;
    property BeforePost;
    property AfterPost;
    property BeforeCancel;
    property AfterCancel;
    property BeforeDelete;
    property AfterDelete;
    property BeforeScroll;
    property AfterScroll;
    property OnCalcFields;
    property OnDeleteError;
    property OnEditError;
    property OnFilterRecord;
    property OnNewRecord;
    property OnPostError;
    property databaseSQL: TcaoDatabaseSQL read FcaoDatabaseSQL write FcaoDatabaseSQL;
    property statementName: WideString read FstatementName write FstatementName;
    property statementText: TStrings read FstatementText write SetLines;
    property cursorName: WideString read FcursorName write FcursorName;
    property openOptions: TcaoOpenOptions read FopenOptions write FopenOptions default caodbRead;
  end;

implementation

{
Initilize properties.
}

constructor TcaoQuery.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FopenOptions := caodbRead;
  FCursorOpen := false;
  FstatementText := TStringList.Create
end;

{
Free SQL statment text.
}

destructor TcaoQuery.Destroy;
begin
  FstatementText.Free;
  inherited Destroy
end;

{
Assign String to statement String list.
}

procedure TcaoQuery.SetLines(Value: TStrings);
begin
  FstatementText.Assign(Value)
end;

{
Handle caoDatabaseSQL component being removed.
}

procedure TcaoQuery.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if Operation = opRemove then
  begin
    if AComponent = FcaoDatabaseSQL then
      FcaoDatabaseSQL := nil
  end
end;

{
Convert Byte array (safe array) to String.
}

function TcaoQuery.VarByteToStr(V: OleVariant): String;
var
  I: Integer;
  Len: Integer;
  S: String;
begin
  Len := VarArrayHighBound(V,1)+1;
  SetLength(S,Len);
  for I := 1 to Len do
    S[I] := Char(Byte(V[I-1]));
  Result := S
end;

{
Convert SmallInt array (safe array) to String.
}

function TcaoQuery.VarSmallIntToStr(V: OleVariant): String;
var
  I: Integer;
  Len: Integer;
  NullInd: SmallInt;
  S: String;
begin
  Len := VarArrayHighBound(V,1)+1;
  SetLength(S,Len*SizeOf(NullInd));
  for I := 0 to Len-1 do
  begin
    NullInd := SmallInt(V[I]);
    Move(NullInd,S[I*SizeOf(NullInd)+1],SizeOf(NullInd))
  end;
  Result := S
end;

{
Get AS/400 specific field definitions and return record length.
}

function TcaoQuery.GetFieldDefs400: Integer;
var
  I: Integer;
  Cols: Integer;
  ColOfs: Integer;
  ColType: SmallInt;
  ColName: WideString;
  ColLength: Integer;
  ColPerc: SmallInt;
  ColScale: SmallInt;
  ColCCSID: SmallInt;
  FieldDef400: TcaoFieldDef;
begin
  ColOfs := 0;
  with FcaoDatabaseSQL do
  begin
    GetResultDataFormat;
    with databaseDataFormat do
    begin
      FFieldDefs400.Clear;
      Cols := columnCount;
      for I := 1 to Cols do
      begin
        FieldDef400 := TcaoFieldDef(FFieldDefs400.Add);
        GetColumnName(I,ColName);
        FieldDef400.ColName := ColName;
        GetColumnType(I,ColType);
        FieldDef400.ColType := ColType;
        GetColumnLength(I,ColLength);
        FieldDef400.ColLength := ColLength;
        GetColumnPrecision(I,ColPerc);
        FieldDef400.ColPerc := ColPerc;
        GetColumnScale(I,ColScale);
        FieldDef400.ColScale := ColScale;
        GetColumnCCSID(I,ColCCSID);
        FieldDef400.ColCCSID := ColCCSID;
        FieldDef400.ColOffset := ColOfs;
        ColOfs := ColOfs+ColLength
      end
    end
  end;
  Result := ColOfs
end;

{
Get result data from last SQL request and copy to cache array.  Return record
count.
}

function TcaoQuery.CacheResultData: Integer;
var
  S: String;
  Ind: String;
  I: Integer;
  IndLen :Integer;
  RecsRead: Integer;
  DataLen: Integer;
  DataBuf: OleVariant;
  IndicatorCount: Integer;
  IndicatorBuf: OleVariant;
begin
  Result := 0;
  with FcaoDatabaseSQL do
  begin
    DataLen := resultDataLength;
    if DataLen > 0 then
    begin
      // Calc records read
      RecsRead := DataLen div FRecNullInfoOfs;
      DataBuf := VarArrayCreate([0,DataLen-1],varByte);
      IndicatorCount := resultIndicatorCount;
      // Calc indicator length per record
      IndLen := (IndicatorCount div RecsRead)*SizeOf(SmallInt);
      IndicatorBuf := VarArrayCreate([0,IndicatorCount-1],varSmallint);
      // Get results from AS/400 server
      GetResultData(DataBuf,DataLen,IndicatorBuf,IndicatorCount);
      // Convert records to a string
      S := VarByteToStr(DataBuf);
      // Convert indicators to a string
      Ind := VarSmallIntToStr(IndicatorBuf);
      // Move converted strings into record cache
      for I := 0 to RecsRead-1 do
      begin
        Move(S[I*FRecNullInfoOfs+1],FRecCache[I][1],FRecNullInfoOfs);
        Move(Ind[I*IndLen+1],FRecCache[I][FRecNullInfoOfs+1],IndLen)
      end;
      Result := RecsRead
    end
  end
end;

{
Move to last record in result set (not EOF).  This can be used to recover from
an EOF condition.
}

procedure TcaoQuery.LastRecord;
begin
  with FcaoDatabaseSQL do
  begin
    defaultBlockCount := 1;
    SetDefaultFetchScrollOpts(cwbdbScrollLast,0);
    Fetch(FcursorName);
    defaultBlockCount := FCacheCount
  end
end;

{
Get next record from cache.  This method is optimized for multiple next reads.
When a buffer is read, the cache cursor (FCurCacheRec) is placed on the first
record in the buffer.  Subsequent next reads will pull from the cache rather
than do a physical read.
}

function TcaoQuery.NextCacheRec: TGetResult;
var
  RecsRead: Integer;
begin
  Result := grOK;
  with FcaoDatabaseSQL do
  begin
    // Both values are -1 when initilized causing first physical read
    if FCurCacheRec >= FLastCacheRec then
    begin
      try
        // Scroll forward 1 record
        SetDefaultFetchScrollOpts(cwbdbScrollRelative,1);
        Fetch(FcursorName);
        RecsRead := CacheResultData;
        FCurRec := FCurRec+RecsRead;
        FLastCacheRec := FCurRec;
        FFirstCacheRec := FLastCacheRec-RecsRead+1;
        FCurCacheRec := FFirstCacheRec
      except
        // EOF reached
        RecsRead := CacheResultData;
        if RecsRead > 0 then
        begin
          FCurRec := FCurRec+RecsRead;
          FLastCacheRec := FCurRec;
          FFirstCacheRec := FLastCacheRec-RecsRead+1;
          FCurCacheRec := FFirstCacheRec
        end
        else
          Result := grEOF;
        // Back up to last record
        LastRecord
      end
    end
    else
      // Get record from cache
      if FCurCacheRec < FLastCacheRec then
        Inc(FCurCacheRec)
      else
        Result := grEOF
  end
end;

{
Get prior record from cache. This method is optimized for multiple prior reads.
When a buffer is read, the cache cursor (FCurCacheRec) is placed on the last
record in the buffer.  Subsequent prior reads will pull from the cache rather
than do a physical read.
}

function TcaoQuery.PriorCacheRec: TGetResult;
var
  RecsRead: Integer;
  Offset: Integer;
begin
  Result := grOK;
  with FcaoDatabaseSQL do
  begin
    if FCurCacheRec <= FFirstCacheRec then
    begin
      try
        if FCurCacheRec > 0 then
        begin
          //  Back up last cache buffer + 1 full cache buffer - 1 record
          Offset := -((FLastCacheRec-FFirstCacheRec)+FCacheCount);
          if FCurRec+Offset >= 0 then
          begin
            SetDefaultFetchScrollOpts(cwbdbScrollRelative,Offset);
            Fetch(FcursorName);
            RecsRead := CacheResultData;
            FCurRec := FCurRec+Offset+RecsRead-1;
            FFirstCacheRec := FCurRec-RecsRead+1;
            FLastCacheRec := FCurRec;
            FCurCacheRec := FLastCacheRec
          end
          else
          begin
            // Read from first record
            Offset := -FCurRec;
            SetDefaultFetchScrollOpts(cwbdbScrollRelative,Offset);
            Fetch(FcursorName);
            RecsRead := CacheResultData;
            FCurCacheRec := FCurCacheRec-1;
            FCurRec := RecsRead-1;
            FFirstCacheRec := FCurRec-RecsRead+1;
            FLastCacheRec := FCurRec
          end
        end
        else
          Result := grBOF
      except
        // This should never execute
        Result := grError
      end
    end
    else
      // Get record from cache
      if FCurCacheRec > FFirstCacheRec then
        Dec(FCurCacheRec)
      else
        Result := grError
  end
end;

{
FopenOptions sets flag.
}

function TcaoQuery.GetCanModify: Boolean;
begin
  Result := not (FopenOptions in [caodbRead])
end;

{
This method is called by TDataSet.Open and also when FieldDefs need to be
updated (usually by the DataSet designer).  Everything which is allocated or
initialized in this method should also be freed or uninitialized in the
InternalClose method.
}

procedure TcaoQuery.InternalOpen;
var
  OpenOpt: SmallInt;
  I: Integer;
  S: String;
begin
  with FcaoDatabaseSQL do
  begin
    if not Active then
      Active := true;
    // Raise exception if block count < 1 block.
    if defaultBlockCount <  1 then
      raise EcaoQuery.Create('TcaoDatabaseSQL.defaultBlockCount must be > 0');
    // Raise exception if cursor not scrollable.
    if defaultScrollableCursor <> caodbCursorScrollable then
      raise EcaoQuery.Create(
      'TcaoDatabaseSQL.defaultScrollableCursor must be caodbCursorScrollable');
    // Set open options from enumerated property
    case FopenOptions of
      caodbRead: OpenOpt := cwbdbRead;
      caodbWrite: OpenOpt := cwbdbWrite;
      caodbUpdate: OpenOpt := cwbdbUpdate;
      caodbDelete: OpenOpt := cwbdbDelete;
      caodbOpenAll: OpenOpt := cwbdbOpenAll
    else
      OpenOpt := cwbdbRead
    end;
    S := '';
    // Build SQL statment from string list
    for I := 0 to FstatementText.Count-1 do
      S := S+FstatementText.Strings[I]+' ';
    // Prepare SQL statement
    PrepareDescribe(FstatementName,Trim(S));
    // Open cursor
    Open(FstatementName,FcursorName,OpenOpt);
    FCursorOpen := true;
    FFieldDefs400 := TCollection.Create(TcaoFieldDef);
    // Initialize our internal physical position.
    // -1 is used to indicate the "crack" before the first record.
    FCurRec := -1;
    // Initialize an offset value to find the null indicators
    FRecNullInfoOfs := GetFieldDefs400;
    // Initialize an offset value to find the TcaoRecInfo in each buffer
    FRecInfoOfs := FRecNullInfoOfs+FFieldDefs400.Count*SizeOf(SmallInt);
    // Calculate the size of the record buffer.
    // Note: This is NOT the same as the RecordSize property which
    // only gets the size of the data in the record buffer
    FRecBufSize := FRecInfoOfs+SizeOf(TcaoRecInfo);
    // Initilize record cache
    FCacheCount := defaultBlockCount;
    SetLength(FRecCache,FCacheCount);
    FCurCacheRec := -1;
    FFirstCacheRec := -1;
    FLastCacheRec := -1;
    // Initilize cache records (without bookmark info)
    for I := 0 to FCacheCount-1 do
      SetLength(FRecCache[I],FRecInfoOfs);
    // Tell TDataSet how big our Bookmarks are (REQUIRED)
    BookmarkSize := SizeOf(Integer);
    // Initialize the FieldDefs
    InternalInitFieldDefs;
    // Create TField components when no persistent fields have been created
    if DefaultFields then
      CreateFields;
    // Bind the TField components to the physical fields
    BindFields(True)
  end
end;

{
Close down objects created in InternalOpen.
}

procedure TcaoQuery.InternalClose;
begin
  with FcaoDatabaseSQL do
  begin
    // Close cursor
    if FCursorOpen then
      Close(FcursorName);
    // Free AS/400 specific field definitions
    FFieldDefs400.Free;
    // Destroy the TField components if no persistent fields
    if DefaultFields then
      DestroyFields;
    // Reset these internal flags
    FCurRec := -1;
    FCurCacheRec := -1;
    FFirstCacheRec := -1;
    FLastCacheRec := -1;
    FCursorOpen := false
  end
end;

{
This property is used while opening the dataset.  It indicates if data is
available even though the current state is still dsInActive.
}

function TcaoQuery.IsCursorOpen: Boolean;
begin
  Result := FCursorOpen
end;

{
Create FieldDefs based on AS/400 field types.
}

procedure TcaoQuery.InternalInitFieldDefs;
var
  I: Integer;
  Cols: Integer;
  FieldDef400: TcaoFieldDef;
begin
  FieldDefs.Clear;
  Cols := FFieldDefs400.Count;
  for I := 1 to Cols do
  begin
    FieldDef400 := TcaoFieldDef(FFieldDefs400.Items[I-1]);
    with FieldDef400 do
    begin
      case ColType of
        cwbdbPCString,
        cwbdbPCVarString:
          TFieldDef.Create(FieldDefs,ColName,ftString,ColLength,false,I);
        // Convert packed and zoned types to native Delphi types based on
        // scale and precision.
        cwbdbPCPacked,
        cwbdbPCZoned:
          begin
            if ColScale > 0 then
              TFieldDef.Create(FieldDefs,ColName,ftFloat,0,false,I)
            else
              if ColPerc < 5 then
                TFieldDef.Create(FieldDefs,ColName,ftSmallInt,0,false,I)
              else
                if ColPerc < 10 then
                  TFieldDef.Create(FieldDefs,ColName,ftInteger,0,false,I)
                else
                  TFieldDef.Create(FieldDefs,ColName,ftFloat,0,false,I)
          end;
        cwbdbPCLong:
          TFieldDef.Create(FieldDefs,ColName,ftInteger,0,false,I);
        cwbdbPCShort:
          TFieldDef.Create(FieldDefs,ColName,ftSmallInt,0,false,I);
        cwbdbPCFloat,
        cwbdbPCDouble:
          TFieldDef.Create(FieldDefs,ColName,ftFloat,0,false,I);
        cwbdbPCGraphic,
        cwbdbPCVargraphic:
          TFieldDef.Create(FieldDefs,ColName,ftGraphic,ColLength,false,I);
        // I'm not exactly sure when these would ever be used
        cwbdbPCNoConversion,
        cwbdbPCInvalidType:
          TFieldDef.Create(FieldDefs,ColName,ftUnknown,0,false,I)
      end
    end
  end
end;

{
This is the exception handler which is called if an exception is raised
while the component is being stream in or streamed out.  In most cases this
should be implemented useing the application exception handler as follows.
}

procedure TcaoQuery.InternalHandleException;
begin
  Application.HandleException(Self)
end;

{
Bookmarks are stored in the record buffer holding the data.  Positioning to a
bookmark just requires finding the offset of the bookmark from the current
record position.  SetDefaultFetchScrollOpts is used with cwbdbScrollRelative
for relative positioning.
}

procedure TcaoQuery.InternalGotoBookmark(Bookmark: Pointer);
var
  RelRec: Integer;
  RecsRead: Integer;
  Offset: Integer;
begin
  with FcaoDatabaseSQL do
  begin
    RelRec := PInteger(Bookmark)^;
    if RelRec > -1 then
    begin
      // See if record exists in cache
      if (RelRec >= FFirstCacheRec) and (RelRec <= FLastCacheRec) then
        FCurCacheRec := RelRec
      else
      begin
        // Calc relative position from bookmark record - current record
        try
          Offset := RelRec-FCurRec;
          SetDefaultFetchScrollOpts(cwbdbScrollRelative,Offset);
          Fetch(FcursorName);
          RecsRead := CacheResultData;
          FCurRec := RelRec+RecsRead-1;
          FFirstCacheRec := RelRec;
          FLastCacheRec := FCurRec;
          FCurCacheRec := FFirstCacheRec
        except
          // EOF reached
          RecsRead := CacheResultData;
          if RecsRead > 0 then
          begin
            FCurRec := RelRec+RecsRead-1;
            FFirstCacheRec := RelRec;
            FLastCacheRec := FCurRec;
            FCurCacheRec := FFirstCacheRec;
            // Back up to last record
            LastRecord
          end
          else
            DatabaseError('Bookmark not found')
        end
      end
    end
    else
      DatabaseError('Bookmark not found')
  end
end;

{
This function does the same thing as InternalGotoBookmark, but it takes a
record buffer as a parameter instead.
}

procedure TcaoQuery.InternalSetToRecord(Buffer: PChar);
begin
  InternalGotoBookmark(@PcaoRecInfo(Buffer+FRecInfoOfs).Bookmark)
end;

{
Bookmark flags are used to indicate if a particular record is the first or last
record in the dataset.  This is necessary for "crack" handling.  If the
bookmark flag is bfBOF or bfEOF then the bookmark is not actually used;
InternalFirst, or InternalLast are called instead by TDataSet.
}

function TcaoQuery.GetBookmarkFlag(Buffer: PChar): TBookmarkFlag;
begin
  Result := PcaoRecInfo(Buffer+FRecInfoOfs).BookmarkFlag
end;

procedure TcaoQuery.SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag);
begin
  PcaoRecInfo(Buffer+FRecInfoOfs).BookmarkFlag := Value
end;

{
These methods provide a way to read and write bookmark data into the record
buffer without actually repositioning the current record.
}

procedure TcaoQuery.GetBookmarkData(Buffer: PChar; Data: Pointer);
begin
  PInteger(Data)^ := PcaoRecInfo(Buffer+FRecInfoOfs).Bookmark
end;

procedure TcaoQuery.SetBookmarkData(Buffer: PChar; Data: Pointer);
begin
  PcaoRecInfo(Buffer+FRecInfoOfs).Bookmark := PInteger(Data)^
end;

{
This method returns the size of just the data in the record buffer.  Do not
confuse this with RecBufSize which also includes any additonal structures
stored in the record buffer (such as TRecInfo).
}

function TcaoQuery.GetRecordSize: Word;
begin
  Result := FRecNullInfoOfs
end;

{
TDataSet calls this method to allocate the record buffer.  Here we use
FRecBufSize which is equal to the size of the data plus the size of the
TRecInfo structure.
}

function TcaoQuery.AllocRecordBuffer: PChar;
begin
  Result := AllocMem(FRecBufSize)
end;

{
Again, TDataSet calls this method to free the record buffer.  Note: Make sure
the value of FRecBufSize does not change before all allocated buffers are
freed.
}

procedure TcaoQuery.FreeRecordBuffer(var Buffer: PChar);
begin
  FreeMem(Buffer)
end;

{
This multi-purpose function does 3 jobs.  It retrieves data for either the
current, the prior, or the next record.  It must return the status
(TGetResult), and raise an exception if DoCheck is True.
}

function TcaoQuery.GetRecord(Buffer: PChar; GetMode: TGetMode;
  DoCheck: Boolean): TGetResult;
begin
  with FcaoDatabaseSQL do
  begin
    try
      Result := grOK;
      case GetMode of
        gmNext:
          Result := NextCacheRec;
        gmPrior:
          Result := PriorCacheRec;
        gmCurrent:
          if FCurRec < 0 then
            Result := grError
      end;
      if Result = grOK then
      begin
        Move(FRecCache[FCurCacheRec-FFirstCacheRec][1],Buffer^,FRecInfoOfs);
        with PcaoRecInfo(Buffer+FRecInfoOfs)^ do
        begin
          BookmarkFlag := bfCurrent;
          Bookmark := Integer(FCurCacheRec)
        end
      end
      else
        if (Result = grError) and DoCheck then
          DatabaseError('No Records')
    except
      Result := grError
    end
  end
end;

{
This routine is called to initialize a record buffer with zero values, but you
might have code to initialize default values or do other things as well.
}

procedure TcaoQuery.InternalInitRecord(Buffer: PChar);
begin
  FillChar(Buffer^,FRecBufSize,0)
end;

{
Copy the data from the record buffer into a field's buffer.  This function, and
SetFieldData, are more complex when supporting calculated fields, filters, and
other more advanced features.  Data is converted from AS/400 format to native
Delphi types.
}

function TcaoQuery.GetFieldData(Field: TField; Buffer: Pointer): Boolean;
var
  ShortVar: SmallInt;
  NullInd: SmallInt;
  IntegerVar: Integer;
  SingleVar: Single;
  DoubleVar: Double;
  PackedStr: String;
  P: Integer;
  EndPos: Integer;
  FieldDef400: TcaoFieldDef;
  S: PChar;
begin
  S := ActiveBuffer;
  // There is one indicator returned for each column of each row in the result
  // set.  Each entry specifes if the corresponding data value is null (-1) or
  // if it experienced a conversion error ('E' or 0x0045).  Otherwise, the
  // value is zero.
  Move(S[FRecNullInfoOfs+(Field.FieldNo-1)*SizeOf(NullInd)],
  NullInd,SizeOf(NullInd));
  Result := (NullInd <> -1);
  if (Result) and (Assigned(Buffer)) then
  begin
    FillChar(Buffer^,Field.DataSize,0);
    FieldDef400 := TcaoFieldDef(FFieldDefs400.Items[Field.FieldNo-1]);
    with FieldDef400 do
    begin
      case ColType of
        cwbdbPCString:
          begin
            SetLength(PackedStr,ColLength);
            Move(S[ColOffset],PackedStr[1],ColLength);
            // Remove any spaces
            PackedStr := Trim(PackedStr);
            Move(PackedStr[1],Buffer^,Length(PackedStr))
          end;
        cwbdbPCVarString,
        cwbdbPCVargraphic:
          begin
            // Get actual length from first two bytes
            Move(S[ColOffset],ShortVar,SizeOf(ShortVar));
            Move(S[ColOffset+2],Buffer^,ShortVar)
          end;
        cwbdbPCPacked:
          begin
            // Handle empty table
            if FCurRec > -1 then
            begin
              PackedStr := '';
              EndPos := ColOffset+ColLength-1;
              for P := ColOffset to EndPos do
              begin
                // Convert left nibble (4 bits)
                PackedStr := PackedStr+Char((Byte(S[P]) and $f0) shr 4+48);
                // Convert right nibble if not last Byte
                if P <> EndPos then
                  PackedStr := PackedStr+Char((Byte(S[P]) and $0f)+48)
              end;
              // Add decimal place holder if needed
              if ColScale > 0 then
                SYSTEM.Insert('.',PackedStr,Length(PackedStr)-ColScale+1);
              // Add sign if needed
              if (Byte(S[EndPos]) and $0f) = $0d then
                PackedStr := '-'+PackedStr;
              // Convert to native Delphi numeric types
              if ColScale > 0 then
              begin
                DoubleVar := StrToFloat(PackedStr);
                Move(DoubleVar,Buffer^,Field.DataSize)
              end
              else
                if ColPerc < 5 then
                begin
                  ShortVar := StrToInt(PackedStr);
                  Move(ShortVar,Buffer^,Field.DataSize)
                end
                else
                  if ColPerc < 10 then
                  begin
                    IntegerVar := StrToInt(PackedStr);
                    Move(IntegerVar,Buffer^,Field.DataSize)
                  end
                  else
                  begin
                    DoubleVar := StrToFloat(PackedStr);
                    Move(DoubleVar,Buffer^,Field.DataSize)
                  end
            end
            else
            begin
              // Null field for empty table
              NullInd := -1;
              Move(NullInd,S[FRecNullInfoOfs+(Field.FieldNo-1)*SizeOf(NullInd)],
              SizeOf(NullInd))
            end
          end;
        cwbdbPCZoned:
          begin
            // Handle empty table
            if FCurRec > -1 then
            begin
              SetLength(PackedStr,ColLength);
              Move(S[ColOffset],PackedStr[1],ColLength);
              // Add sign if needed
              if (Byte(PackedStr[ColLength]) and $40) = $40 then
              begin
                Byte(PackedStr[ColLength]) :=
                Byte(PackedStr[ColLength]) and $3F;
                PackedStr := '-'+PackedStr
              end;
              // Add decimal place holder if needed
              if ColScale > 0 then
                SYSTEM.Insert('.',PackedStr,Length(PackedStr)-ColScale+1);
              // Convert to native Delphi numeric types
              if ColScale > 0 then
              begin
                DoubleVar := StrToFloat(PackedStr);
                Move(DoubleVar,Buffer^,Field.DataSize)
              end
              else
                if ColPerc < 5 then
                begin
                  ShortVar := StrToInt(PackedStr);
                  Move(ShortVar,Buffer^,Field.DataSize)
                end
                else
                  if ColPerc < 10 then
                  begin
                    IntegerVar := StrToInt(PackedStr);
                    Move(IntegerVar,Buffer^,Field.DataSize)
                  end
                  else
                  begin
                    DoubleVar := StrToFloat(PackedStr);
                    Move(DoubleVar,Buffer^,Field.DataSize)
                  end
            end
            else
            begin
              // Null field for empty table
              NullInd := -1;
              Move(NullInd,S[FRecNullInfoOfs+(Field.FieldNo-1)*SizeOf(NullInd)],
              SizeOf(NullInd))
            end
          end;
        cwbdbPCFloat:
          begin
            Move(S[ColOffset],SingleVar,SizeOf(SingleVar));
            DoubleVar := SingleVar;
            Move(DoubleVar,Buffer^,Field.DataSize)
          end;
        cwbdbPCDouble,
        cwbdbPCLong,
        cwbdbPCShort,
        cwbdbPCGraphic,
        cwbdbPCNoConversion,
        cwbdbPCInvalidType:
          Move(S[ColOffset],Buffer^,Field.DataSize)
      end
    end
  end
end;

{
Copy the data from the field buffer into a record's buffer.  Data is converted
to from Delphi types to AS/400 format.
}

procedure TcaoQuery.SetFieldData(Field: TField; Buffer: Pointer);
var
  ShortVar: SmallInt;
  NullInd: SmallInt;
  SingleVar: Single;
  DoubleVar: Double;
  PackedStr: String;
  FormatStr: String;
  NumStr: String;
  LowNibble: Boolean;
  P: Integer;
  I: Integer;
  EndPos: Integer;
  FieldDef400: TcaoFieldDef;
  S: PChar;
begin
  S := ActiveBuffer;
  FieldDef400 := TcaoFieldDef(FFieldDefs400.Items[Field.FieldNo-1]);
  with FieldDef400 do
  begin
    // Zero out active buffer
    FillChar(S[ColOffset],ColLength,0);
    if not Field.IsNull then
    begin
      NullInd := 0;
      case ColType of
        cwbdbPCString:
          begin
            Move(Buffer^,S[ColOffset],Length(PChar(Buffer)))
          end;
        cwbdbPCVarString,
        cwbdbPCVargraphic:
          begin
            // Get actual length from first two bytes
            ShortVar := Length(PChar(Buffer));
            Move(ShortVar,S[ColOffset],SizeOf(ShortVar));
            Move(Buffer^,S[ColOffset+2],ShortVar)
          end;
        cwbdbPCPacked:
          begin
            // Convert to native AS/400 numeric types
            if ColScale > 0 then
            begin
              FormatStr := '%'+IntToStr(ColPerc)+'.'+IntToStr(ColScale)+'f';
              NumStr := Format(FormatStr,[Double(Buffer^)])
            end
            else
              if ColPerc < 5 then
              begin
                NumStr := IntToStr(Short(Buffer^))
              end
              else
                if ColPerc < 10 then
                begin
                  NumStr := IntToStr(Integer(Buffer^))
                end
                else
                begin
                  FormatStr := '%'+IntToStr(ColPerc)+'.0f';
                  NumStr := Format(FormatStr,[Double(Buffer^)])
                end;
            EndPos := Length(NumStr);
            SetLength(PackedStr,ColLength);
            FillChar(PackedStr[1],ColLength,0);
            I := ColLength;
            // Set sign
            if Pos('-',NumStr) > 0 then
              PackedStr[I] := #$0d
            else
              PackedStr[I] := #$0f;
            // Start off with high nibble since sign is in low nibble
            LowNibble := False;
            for P := EndPos downto 1 do
            begin
              case NumStr[P] of
              '0'..'9':
                if LowNibble then
                begin
                  PackedStr[I] := Char(Byte(NumStr[P])-48);
                  LowNibble := false
                end
                else
                begin
                  PackedStr[I] := Char(Byte(PackedStr[I]) or
                  ((Byte(NumStr[P])-48) shl 4));
                  Dec(I);
                  LowNibble := true
                end
              end
            end;
            Move(PackedStr[1],S[ColOffset],ColLength)
          end;
        cwbdbPCZoned:
          begin
            // Convert to native AS/400 numeric types
            if ColScale > 0 then
            begin
              FormatStr := '%'+IntToStr(ColPerc)+'.'+IntToStr(ColScale)+'f';
              NumStr := Format(FormatStr,[Double(Buffer^)])
            end
            else
              if ColPerc < 5 then
              begin
                NumStr := IntToStr(Short(Buffer^))
              end
              else
                if ColPerc < 10 then
                begin
                  NumStr := IntToStr(Integer(Buffer^))
                end
                else
                begin
                  FormatStr := '%'+IntToStr(ColPerc)+'.0f';
                  NumStr := Format(FormatStr,[Double(Buffer^)])
                end;
            EndPos := Length(NumStr);
            SetLength(PackedStr,ColLength);
            FillChar(PackedStr[1],ColLength,'0');
            I := ColLength;
            for P := EndPos downto 1 do
            begin
              case NumStr[P] of
              '0'..'9':
                begin
                  PackedStr[I] := NumStr[P];
                  Dec(I)
                end
              end
            end;
            // Set sign
            if Pos('-',NumStr) > 0 then
              Byte(PackedStr[ColLength]) :=
              Byte(PackedStr[ColLength]) or $40;
            Move(PackedStr[1],S[ColOffset],ColLength)
          end;
        cwbdbPCFloat:
          begin
            Move(Buffer^,DoubleVar,SizeOf(DoubleVar));
            SingleVar := DoubleVar;
            Move(SingleVar,S[ColOffset],SizeOf(SingleVar))
          end;
        cwbdbPCDouble,
        cwbdbPCLong,
        cwbdbPCShort,
        cwbdbPCGraphic,
        cwbdbPCNoConversion,
        cwbdbPCInvalidType:
          Move(Buffer^,S[ColOffset],Field.DataSize)
      end
    end
    else
      NullInd := -1;
    // There is one indicator returned for each column of each row in the result
    // set.  Each entry specifes if the corresponding data value is null (-1) or
    // if it experienced a conversion error ('E' or 0x0045).  Otherwise, the
    // value is zero.
    Move(NullInd,S[FRecNullInfoOfs+(Field.FieldNo-1)*SizeOf(NullInd)],
    SizeOf(NullInd))
  end;
  DataEvent(deFieldChange,Longint(Field))
end;

{
This method is called by TDataSet.First.  Crack behavior is required.  That is
you must position to a special place *before* the first record.  Otherwise,
you will actually end up on the second record after Resync is called.
}

procedure TcaoQuery.InternalFirst;
begin
  with FcaoDatabaseSQL do
  begin
    SetDefaultFetchScrollOpts(cwbdbScrollBeforeFirst,0);
    Fetch(FcursorName)
  end;
  FCurRec := -1;
  FCurCacheRec := -1;
  FFirstCacheRec := -1;
  FLastCacheRec := -1
end;

{
Again, we position to the crack *after* the last record here.
}

procedure TcaoQuery.InternalLast;
begin
  while NextCacheRec = grOK do;
  Inc(FCurCacheRec)
end;

{
This method is called by TDataSet.Post.  This needs to be implemented to update
and insert records.  SetFieldData has already been implemented.
}

procedure TcaoQuery.InternalPost;
begin
  // For inserts, use UPDATE SQL command
  if State = dsEdit then
  begin
  end
end;

{
Execute SQL statement.  Used for non-cursor oriented SQL like CREATE TABLE,
etc.
}

procedure TcaoQuery.ExecSQL;
var
  I: Integer;
  S: String;
begin
  with FcaoDatabaseSQL do
  begin
    if not Active then
      Active := true;
    S := '';
    // Build SQL statment from string list
    for I := 0 to FstatementText.Count-1 do
      S := S+FstatementText.Strings[I]+' ';
    // Prepare and execute SQL statement
    ExecuteImmediate(FstatementName,S,FcursorName)
  end
end;

end.
